from typing import Any

from ..helpers import parse_uint24_be
from .device import SwitchbotDevice, SwitchbotOperationError

COMMAND_SET_TIME_OFFSET = "570f680506"
COMMAND_GET_TIME_OFFSET = "570f690506"
MAX_TIME_OFFSET = (1 << 24) - 1

COMMAND_GET_DEVICE_DATETIME = "570f6901"
COMMAND_SET_DEVICE_DATETIME = "57000503"
COMMAND_SET_DISPLAY_FORMAT = "570f680505"


class SwitchbotMeterProCO2(SwitchbotDevice):
    """API to control Switchbot Meter Pro CO2."""

    async def get_time_offset(self) -> int:
        """
        Get the current display time offset from the device.

        Returns:
            int: The time offset in seconds. Max 24 bits.

        """
        # Response Format: 5 bytes, where
        # - byte 0: "01" (success)
        # - byte 1: "00" (plus offset) or "80" (minus offset)
        # - bytes 2-4: int24, number of seconds to offset.
        # Example response: 01-80-00-10-00 -> subtract 4096 seconds.
        result = await self._send_command(COMMAND_GET_TIME_OFFSET)
        result = self._validate_result("get_time_offset", result, min_length=5)

        is_negative = bool(result[1] & 0b10000000)
        offset = parse_uint24_be(result, 2)
        return -offset if is_negative else offset

    async def set_time_offset(self, offset_seconds: int) -> None:
        """
        Set the display time offset on the device. This is what happens when
        you adjust display time in the Switchbot app. The displayed time is
        calculated as the internal device time (usually comes from the factory
        settings or set by the Switchbot app upon syncing) + offset. The offset
        is provided in seconds and can be positive or negative.

        Args:
            offset_seconds (int): 2^24 maximum, can be negative.

        """
        abs_offset = abs(offset_seconds)
        if abs_offset > MAX_TIME_OFFSET:
            raise SwitchbotOperationError(
                f"{self.name}: Requested to set_time_offset of {offset_seconds} seconds, allowed +-{MAX_TIME_OFFSET} max."
            )

        sign_byte = "80" if offset_seconds < 0 else "00"

        # Example: 57-0f-68-05-06-80-00-10-00 -> subtract 4096 seconds.
        payload = f"{COMMAND_SET_TIME_OFFSET}{sign_byte}{abs_offset:06x}"
        result = await self._send_command(payload)
        self._validate_result("set_time_offset", result)

    async def get_datetime(self) -> dict[str, Any]:
        """
        Get the current device time and settings as it is displayed. Contains
        a time offset, if any was applied (see set_time_offset).
        Doesn't include the current time zone.

        Returns:
            dict: Dictionary containing:
                - 12h_mode (bool): True if 12h mode, False if 24h mode.
                - year (int)
                - month (int)
                - day (int)
                - hour (int)
                - minute (int)
                - second (int)

        """
        # Response Format: 13 bytes, where
        # - byte 0: "01" (success)
        # - bytes 1-4: temperature, ignored here.
        # - byte 5: time display format:
        #   - "80" - 12h (am/pm)
        #   - "00" - 24h
        # - bytes 6-12: yyyy-MM-dd-hh-mm-ss
        # Example: 01-e4-02-94-23-00-07-e9-0c-1e-08-37-01 contains
        # "year 2025, 30 December, 08:55:01, displayed in 24h format".
        result = await self._send_command(COMMAND_GET_DEVICE_DATETIME)
        result = self._validate_result("get_datetime", result, min_length=13)
        return {
            # Whether the time is displayed in 12h(am/pm) or 24h mode.
            "12h_mode": bool(result[5] & 0b10000000),
            "year": (result[6] << 8) + result[7],
            "month": result[8],
            "day": result[9],
            "hour": result[10],
            "minute": result[11],
            "second": result[12],
        }

    async def set_datetime(
        self, timestamp: int, utc_offset_hours: int = 0, utc_offset_minutes: int = 0
    ) -> None:
        """
        Set the device internal time and timezone. Similar to how the
        Switchbot app does it upon syncing with the device.
        Pay attention to calculating UTC offset hours and minutes, see
        examples below.

        Args:
            timestamp (int): Unix timestamp in seconds.
            utc_offset_hours (int): UTC offset in hours, floor()'ed,
                within [-12; 14] range.
                Examples: -5 for UTC-05:00, -6 for UTC-05:30,
                5 for UTC+05:00, 5 for UTC+5:30.
            utc_offset_minutes (int): UTC offset minutes component, always
                positive, complements utc_offset_hours.
                Examples: 45 for UTC+05:45, 15 for UTC-5:45.

        """
        if not (-12 <= utc_offset_hours <= 14):
            raise SwitchbotOperationError(
                f"{self.name}: utc_offset_hours must be between -12 and +14 inclusive, got {utc_offset_hours}"
            )
        if not (0 <= utc_offset_minutes < 60):
            raise SwitchbotOperationError(
                f"{self.name}: utc_offset_minutes must be between 0 and 59 inclusive, got {utc_offset_minutes}"
            )

        # The device doesn't automatically add offset minutes, it expects them
        # to come as a part of the timestamp.
        adjusted_timestamp = timestamp + utc_offset_minutes * 60

        # The timezone is encoded as 1 byte, where 00 stands for UTC-12.
        # TZ with minute offset gets floor()ed: 4:30 yields 4, -4:30 yields -5.
        utc_byte = utc_offset_hours + 12

        payload = (
            f"{COMMAND_SET_DEVICE_DATETIME}{utc_byte:02x}"
            f"{adjusted_timestamp:016x}{utc_offset_minutes:02x}"
        )

        result = await self._send_command(payload)
        self._validate_result("set_datetime", result)

    async def set_time_display_format(self, is_12h_mode: bool = False) -> None:
        """
        Set the time display format on the device: 12h(AM/PM) or 24h.

        Args:
            is_12h_mode (bool): True for 12h (AM/PM) mode, False for 24h mode.

        """
        mode_byte = "80" if is_12h_mode else "00"
        payload = f"{COMMAND_SET_DISPLAY_FORMAT}{mode_byte}"
        result = await self._send_command(payload)
        self._validate_result("set_time_display_format", result)

    def _validate_result(
        self, op_name: str, result: bytes | None, min_length: int | None = None
    ) -> bytes:
        if not self._check_command_result(result, 0, {1}):
            raise SwitchbotOperationError(
                f"{self.name}: Unexpected response code for {op_name} (result={result.hex() if result else 'None'} rssi={self.rssi})"
            )
        assert result is not None
        if min_length is not None and len(result) < min_length:
            raise SwitchbotOperationError(
                f"{self.name}: Unexpected response len for {op_name}, wanted at least {min_length} (result={result.hex() if result else 'None'} rssi={self.rssi})"
            )
        return result



"""Advertisement data parser for climate panel devices."""

import logging

_LOGGER = logging.getLogger(__name__)


def process_climate_panel(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, bool | int | str]:
    """Process Climate Panel data."""
    if mfr_data is None or len(mfr_data) < 16:
        return {}

    seq_number = mfr_data[6]
    isOn = bool(mfr_data[7] & 0x80)
    battery = mfr_data[7] & 0x7F
    humidity_alarm = (mfr_data[8] >> 6) & 0x03
    temp_alarm = (mfr_data[8] >> 4) & 0x03

    temp_decimal = mfr_data[8] & 0x0F
    temp_sign = 1 if (mfr_data[9] & 0x80) else -1
    temp_int = mfr_data[9] & 0x7F
    temperature = temp_sign * (temp_int + temp_decimal / 10)

    humidity = mfr_data[10] & 0x7F

    pir_state = bool(mfr_data[15] & 0x80)
    is_light = ((mfr_data[15] >> 1) & 0x03) == 0x02

    result = {
        "sequence_number": seq_number,
        "isOn": isOn,
        "battery": battery,
        "temperature": temperature,
        "humidity": humidity,
        "temp_alarm": temp_alarm,
        "humidity_alarm": humidity_alarm,
        "motion_detected": pir_state,
        "is_light": is_light,
    }

    _LOGGER.debug(
        "Processed climate panel mfr data: %s, result: %s", mfr_data.hex(), result
    )
    return result



"""Representation of climate-related constants."""

from enum import Enum


class ClimateMode(Enum):
    """Climate Modes."""

    OFF = 0
    HEAT = 1
    COOL = 2
    HEAT_COOL = 3
    AUTO = 4
    DRY = 5
    FAN_ONLY = 6


class ClimateAction(Enum):
    """Climate Actions."""

    OFF = 0
    IDLE = 1
    HEATING = 2


class SmartThermostatRadiatorMode(Enum):
    """Smart Thermostat Radiator Modes."""

    SCHEDULE = 0
    MANUAL = 1
    OFF = 2
    ECO = 3
    COMFORT = 4
    BOOST = 5

    @property
    def lname(self) -> str:
        return self.name.lower()

    @classmethod
    def get_modes(cls) -> list[str]:
        return [mode.lname for mode in cls]

    @classmethod
    def get_mode_name(cls, mode_value: int) -> str:
        return cls(mode_value).lname

    @classmethod
    def get_valid_modes(cls) -> list[str]:
        return [mode.lname for mode in cls if mode != cls.OFF]



"""Advertisement data parser for presence sensor devices."""

from __future__ import annotations

import logging

from ..const.presence_sensor import BATTERY_LEVEL_MAP

_LOGGER = logging.getLogger(__name__)


def process_presence_sensor(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, bool | int | str]:
    """Process Presence Sensor data."""
    if mfr_data is None or len(mfr_data) < 12:
        return {}

    seq_number = mfr_data[6]
    adaptive_state = bool(mfr_data[7] & 0x80)
    motion_detected = bool(mfr_data[7] & 0x40)
    battery_bits = (mfr_data[7] >> 2) & 0x03
    battery_range = BATTERY_LEVEL_MAP.get(battery_bits, "Unknown")
    duration = (mfr_data[8] << 8) + mfr_data[9]
    trigger_flag = mfr_data[10]
    led_state = bool(mfr_data[11] & 0x80)
    light_level = mfr_data[11] & 0x1F

    result = {
        "sequence_number": seq_number,
        "adaptive_state": adaptive_state,
        "motion_detected": motion_detected,
        "battery_range": battery_range,
        "duration": duration,
        "trigger_flag": trigger_flag,
        "led_state": led_state,
        "lightLevel": light_level,
    }

    if data:
        battery = data[2] & 0x7F
        result["battery"] = battery

    _LOGGER.debug(
        "Processed presence sensor mfr data: %s, result: %s", mfr_data.hex(), result
    )
    return result



"""Test ColorMode imports for backward compatibility."""


def test_colormode_import_from_main_module():
    """Test that ColorMode can be imported from the main switchbot module."""
    from switchbot import ColorMode  # noqa: PLC0415

    # Verify it's the enum we expect
    assert hasattr(ColorMode, "OFF")
    assert hasattr(ColorMode, "COLOR_TEMP")
    assert hasattr(ColorMode, "RGB")
    assert hasattr(ColorMode, "EFFECT")

    # Verify the values
    assert ColorMode.OFF.value == 0
    assert ColorMode.COLOR_TEMP.value == 1
    assert ColorMode.RGB.value == 2
    assert ColorMode.EFFECT.value == 3


def test_colormode_import_from_device_module():
    """Test that ColorMode can be imported from switchbot.devices.device for backward compatibility."""
    from switchbot.devices.device import ColorMode  # noqa: PLC0415

    # Verify it's the enum we expect
    assert hasattr(ColorMode, "OFF")
    assert hasattr(ColorMode, "COLOR_TEMP")
    assert hasattr(ColorMode, "RGB")
    assert hasattr(ColorMode, "EFFECT")

    # Verify the values
    assert ColorMode.OFF.value == 0
    assert ColorMode.COLOR_TEMP.value == 1
    assert ColorMode.RGB.value == 2
    assert ColorMode.EFFECT.value == 3


def test_colormode_import_from_const():
    """Test that ColorMode can be imported from switchbot.const."""
    from switchbot.const import ColorMode  # noqa: PLC0415

    # Verify it's the enum we expect
    assert hasattr(ColorMode, "OFF")
    assert hasattr(ColorMode, "COLOR_TEMP")
    assert hasattr(ColorMode, "RGB")
    assert hasattr(ColorMode, "EFFECT")

    # Verify the values
    assert ColorMode.OFF.value == 0
    assert ColorMode.COLOR_TEMP.value == 1
    assert ColorMode.RGB.value == 2
    assert ColorMode.EFFECT.value == 3


def test_colormode_import_from_const_light():
    """Test that ColorMode can be imported from switchbot.const.light."""
    from switchbot.const.light import ColorMode  # noqa: PLC0415

    # Verify it's the enum we expect
    assert hasattr(ColorMode, "OFF")
    assert hasattr(ColorMode, "COLOR_TEMP")
    assert hasattr(ColorMode, "RGB")
    assert hasattr(ColorMode, "EFFECT")

    # Verify the values
    assert ColorMode.OFF.value == 0
    assert ColorMode.COLOR_TEMP.value == 1
    assert ColorMode.RGB.value == 2
    assert ColorMode.EFFECT.value == 3


def test_all_colormode_imports_are_same_object():
    """Test that all ColorMode imports reference the same enum object."""
    from switchbot import ColorMode as ColorMode1  # noqa: PLC0415
    from switchbot.const import ColorMode as ColorMode3  # noqa: PLC0415
    from switchbot.const.light import ColorMode as ColorMode4  # noqa: PLC0415
    from switchbot.devices.device import ColorMode as ColorMode2  # noqa: PLC0415

    # They should all be the exact same object
    assert ColorMode1 is ColorMode2
    assert ColorMode2 is ColorMode3
    assert ColorMode3 is ColorMode4

    # And their members should be the same
    assert ColorMode1.OFF is ColorMode2.OFF
    assert ColorMode1.COLOR_TEMP is ColorMode3.COLOR_TEMP
    assert ColorMode1.RGB is ColorMode4.RGB
    assert ColorMode1.EFFECT is ColorMode2.EFFECT



from __future__ import annotations

from enum import Enum


class FanMode(Enum):
    NORMAL = 1
    NATURAL = 2
    SLEEP = 3
    BABY = 4

    @classmethod
    def get_modes(cls) -> list[str]:
        return [mode.name.lower() for mode in cls]


class StandingFanMode(Enum):
    NORMAL = 1
    NATURAL = 2
    SLEEP = 3
    BABY = 4
    CUSTOM_NATURAL = 5

    @classmethod
    def get_modes(cls) -> list[str]:
        return [mode.name.lower() for mode in cls]


class NightLightState(Enum):
    """Standing Fan night-light command values."""

    LEVEL_1 = 1
    LEVEL_2 = 2
    OFF = 3


class HorizontalOscillationAngle(Enum):
    """
    Horizontal oscillation angle command values.

    For the horizontal axis the device byte is the same as the
    user-facing angle in degrees.
    """

    ANGLE_30 = 30
    ANGLE_60 = 60
    ANGLE_90 = 90


class VerticalOscillationAngle(Enum):
    """
    Vertical oscillation angle command values.

    The Standing Fan uses a different byte encoding on the vertical axis
    than on the horizontal one. Byte 0x5A (decimal 90) is interpreted as
    an axis halt, so 90° maps to byte 0x5F (95). 30° and 60° match their
    degree values.
    """

    ANGLE_30 = 30
    ANGLE_60 = 60
    ANGLE_90 = 95



"""Library to handle connection with Switchbot."""

from __future__ import annotations

from typing import Any

from .device import SwitchbotSequenceDevice, update_after_operation

COMMAND_CLEAN_UP = {
    1: "570F5A00FFFF7001",
    2: "5A400101010126",
}
COMMAND_RETURN_DOCK = {
    1: "570F5A00FFFF7002",
    2: "5A400101010225",
}


class SwitchbotVacuum(SwitchbotSequenceDevice):
    """Representation of a Switchbot Vacuum."""

    @update_after_operation
    async def clean_up(self, protocol_version: int) -> bool:
        """Send command to perform a spot clean-up."""
        return await self._send_command(COMMAND_CLEAN_UP[protocol_version])

    @update_after_operation
    async def return_to_dock(self, protocol_version: int) -> bool:
        """Send command to return the dock."""
        return await self._send_command(COMMAND_RETURN_DOCK[protocol_version])

    async def get_basic_info(self) -> dict[str, Any] | None:
        """Only support get the ble version through the command."""
        if not (_data := await self._get_basic_info()):
            return None
        return {
            "firmware": _data[2],
        }

    def get_soc_version(self) -> str:
        """Return device soc version."""
        return self._get_adv_value("soc_version")

    def get_last_step(self) -> int:
        """Return device last step after network configuration."""
        return self._get_adv_value("step")

    def get_mqtt_connnect_status(self) -> bool:
        """Return device mqtt connect status."""
        return self._get_adv_value("mqtt_connected")

    def get_battery(self) -> int:
        """Return device battery."""
        return self._get_adv_value("battery")

    def get_work_status(self) -> int:
        """Return device work status."""
        return self._get_adv_value("work_status")

    def get_dustbin_bound_status(self) -> bool:
        """Return the dustbin bound status"""
        return self._get_adv_value("dustbin_bound")

    def get_dustbin_connnected_status(self) -> bool:
        """Return the dustbin connected status"""
        return self._get_adv_value("dusbin_connected")

    def get_network_connected_status(self) -> bool:
        """Return the network connected status"""
        return self._get_adv_value("network_connected")



"""Art Frame advertisement data parser."""

import logging

_LOGGER = logging.getLogger(__name__)


def process_art_frame(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, bool | int | str]:
    """Process Art Frame data."""
    if mfr_data is None or len(mfr_data) < 10:
        return {}

    _seq_num = mfr_data[6]
    battery_charging = bool(mfr_data[7] & 0x80)
    battery = mfr_data[7] & 0x7F
    image_index = mfr_data[8]
    display_size = (mfr_data[9] >> 4) & 0x0F
    display_mode = (mfr_data[9] >> 3) & 0x01
    last_network_status = (mfr_data[9] >> 2) & 0x01

    result = {
        "sequence_number": _seq_num,
        "battery_charging": battery_charging,
        "battery": battery,
        "image_index": image_index,
        "display_size": display_size,
        "display_mode": display_mode,
        "last_network_status": last_network_status,
    }

    _LOGGER.debug("Art Frame mfr data: %s, result: %s", mfr_data.hex(), result)

    return result



"""Library to handle connection with Switchbot."""

from __future__ import annotations

from bleak_retry_connector import (
    close_stale_connections,
    close_stale_connections_by_address,
    get_device,
)

from .adv_parser import SwitchbotSupportedType, parse_advertisement_data
from .const import (
    AirPurifierMode,
    AirQualityLevel,
    BulbColorMode,
    CeilingLightColorMode,
    ClimateAction,
    ClimateMode,
    ColorMode,
    FanMode,
    HorizontalOscillationAngle,
    HumidifierAction,
    HumidifierMode,
    HumidifierWaterLevel,
    LockStatus,
    NightLightState,
    SmartThermostatRadiatorMode,
    StandingFanMode,
    StripLightColorMode,
    SwitchbotAccountConnectionError,
    SwitchbotApiError,
    SwitchbotAuthenticationError,
    SwitchbotModel,
    VerticalOscillationAngle,
)
from .devices.air_purifier import SwitchbotAirPurifier
from .devices.art_frame import SwitchbotArtFrame
from .devices.base_light import SwitchbotBaseLight
from .devices.blind_tilt import SwitchbotBlindTilt
from .devices.bot import Switchbot
from .devices.bulb import SwitchbotBulb
from .devices.ceiling_light import SwitchbotCeilingLight
from .devices.curtain import SwitchbotCurtain
from .devices.device import (
    SwitchbotDevice,
    SwitchbotEncryptedDevice,
    SwitchbotOperationError,
    fetch_cloud_devices,
)
from .devices.evaporative_humidifier import SwitchbotEvaporativeHumidifier
from .devices.fan import SwitchbotFan, SwitchbotStandingFan
from .devices.humidifier import SwitchbotHumidifier
from .devices.keypad_vision import SwitchbotKeypadVision
from .devices.light_strip import (
    SwitchbotCandleWarmerLamp,
    SwitchbotLightStrip,
    SwitchbotPermanentOutdoorLight,
    SwitchbotRgbicLight,
    SwitchbotRgbicNeonLight,
    SwitchbotRgbicwwCeilingLight,
    SwitchbotStripLight3,
)
from .devices.lock import SwitchbotLock
from .devices.meter_pro import SwitchbotMeterProCO2
from .devices.plug import SwitchbotPlugMini
from .devices.relay_switch import (
    SwitchbotGarageDoorOpener,
    SwitchbotRelaySwitch,
    SwitchbotRelaySwitch2PM,
)
from .devices.roller_shade import SwitchbotRollerShade
from .devices.smart_thermostat_radiator import SwitchbotSmartThermostatRadiator
from .devices.vacuum import SwitchbotVacuum
from .discovery import GetSwitchbotDevices
from .models import SwitchBotAdvertisement

__all__ = [
    "AirPurifierMode",
    "AirQualityLevel",
    "BulbColorMode",
    "CeilingLightColorMode",
    "ClimateAction",
    "ClimateMode",
    "ColorMode",
    "FanMode",
    "GetSwitchbotDevices",
    "HorizontalOscillationAngle",
    "HumidifierAction",
    "HumidifierMode",
    "HumidifierWaterLevel",
    "LockStatus",
    "NightLightState",
    "SmartThermostatRadiatorMode",
    "StandingFanMode",
    "StripLightColorMode",
    "SwitchBotAdvertisement",
    "Switchbot",
    "Switchbot",
    "SwitchbotAccountConnectionError",
    "SwitchbotAirPurifier",
    "SwitchbotApiError",
    "SwitchbotArtFrame",
    "SwitchbotAuthenticationError",
    "SwitchbotBaseLight",
    "SwitchbotBlindTilt",
    "SwitchbotBulb",
    "SwitchbotCandleWarmerLamp",
    "SwitchbotCeilingLight",
    "SwitchbotCurtain",
    "SwitchbotDevice",
    "SwitchbotEncryptedDevice",
    "SwitchbotEvaporativeHumidifier",
    "SwitchbotFan",
    "SwitchbotGarageDoorOpener",
    "SwitchbotHumidifier",
    "SwitchbotKeypadVision",
    "SwitchbotLightStrip",
    "SwitchbotLock",
    "SwitchbotMeterProCO2",
    "SwitchbotModel",
    "SwitchbotModel",
    "SwitchbotOperationError",
    "SwitchbotPermanentOutdoorLight",
    "SwitchbotPlugMini",
    "SwitchbotPlugMini",
    "SwitchbotRelaySwitch",
    "SwitchbotRelaySwitch2PM",
    "SwitchbotRgbicLight",
    "SwitchbotRgbicNeonLight",
    "SwitchbotRgbicwwCeilingLight",
    "SwitchbotRollerShade",
    "SwitchbotSmartThermostatRadiator",
    "SwitchbotStandingFan",
    "SwitchbotStripLight3",
    "SwitchbotSupportedType",
    "SwitchbotSupportedType",
    "SwitchbotVacuum",
    "VerticalOscillationAngle",
    "close_stale_connections",
    "close_stale_connections_by_address",
    "fetch_cloud_devices",
    "get_device",
    "parse_advertisement_data",
]



"""Library to handle connection with Switchbot."""

from __future__ import annotations

import logging
from enum import Enum
from typing import Any, ClassVar

from ..const.fan import (
    FanMode,
    HorizontalOscillationAngle,
    NightLightState,
    StandingFanMode,
    VerticalOscillationAngle,
)
from .device import (
    DEVICE_GET_BASIC_SETTINGS_KEY,
    SwitchbotSequenceDevice,
    update_after_operation,
)

_LOGGER = logging.getLogger(__name__)


COMMAND_HEAD = "570f41"
# Circulator Fan (single-axis): start/stop oscillation with V kept unchanged.
# These also serve as the explicit horizontal-only commands since the byte
# layout is identical.
COMMAND_START_OSCILLATION = f"{COMMAND_HEAD}020101ff"
COMMAND_STOP_OSCILLATION = f"{COMMAND_HEAD}020102ff"
COMMAND_START_HORIZONTAL_OSCILLATION = COMMAND_START_OSCILLATION
COMMAND_STOP_HORIZONTAL_OSCILLATION = COMMAND_STOP_OSCILLATION
COMMAND_START_VERTICAL_OSCILLATION = f"{COMMAND_HEAD}0201ff01"  # H keep, V start
COMMAND_STOP_VERTICAL_OSCILLATION = f"{COMMAND_HEAD}0201ff02"  # H keep, V stop
# Standing Fan (dual-axis): start/stop both axes at once.
COMMAND_START_OSCILLATION_ALL_AXES = f"{COMMAND_HEAD}02010101"
COMMAND_STOP_OSCILLATION_ALL_AXES = f"{COMMAND_HEAD}02010202"
COMMAND_SET_OSCILLATION_PARAMS = f"{COMMAND_HEAD}0202"  # +angles
COMMAND_SET_NIGHT_LIGHT = f"{COMMAND_HEAD}0502"  # +state
# Standing Fan (FAN2) extra controls.
COMMAND_SET_DISPLAY_LIGHT = f"{COMMAND_HEAD}0501"  # +state + FFFF (front LED display)
COMMAND_SET_SOUND = f"{COMMAND_HEAD}0601"  # +level (64 on / 00 off)
COMMAND_SET_AUTO_RECENTER = f"{COMMAND_HEAD}0205"  # +both axes (0101 on / 0202 off)
COMMAND_SET_CHILD_LOCK = f"{COMMAND_HEAD}07"  # +state (01 on / 02 off)
COMMAND_SET_MODE = {
    FanMode.NORMAL.name.lower(): f"{COMMAND_HEAD}030101ff",
    FanMode.NATURAL.name.lower(): f"{COMMAND_HEAD}030102ff",
    FanMode.SLEEP.name.lower(): f"{COMMAND_HEAD}030103",
    FanMode.BABY.name.lower(): f"{COMMAND_HEAD}030104",
}
COMMAND_SET_STANDING_FAN_MODE = {
    **COMMAND_SET_MODE,
    StandingFanMode.CUSTOM_NATURAL.name.lower(): f"{COMMAND_HEAD}030105",
}
COMMAND_SET_PERCENTAGE = f"{COMMAND_HEAD}0302"  #  +speed
COMMAND_GET_BASIC_INFO = "570f428102"


class SwitchbotFan(SwitchbotSequenceDevice):
    """Representation of a Switchbot Circulator Fan."""

    _turn_on_command = f"{COMMAND_HEAD}0101"
    _turn_off_command = f"{COMMAND_HEAD}0102"
    _mode_enum: ClassVar[type[Enum]] = FanMode
    _command_set_mode: ClassVar[dict[str, str]] = COMMAND_SET_MODE
    _command_start_oscillation: ClassVar[str] = COMMAND_START_OSCILLATION
    _command_stop_oscillation: ClassVar[str] = COMMAND_STOP_OSCILLATION

    async def get_basic_info(self) -> dict[str, Any] | None:
        """Get device basic settings."""
        if not (_data := await self._get_basic_info(COMMAND_GET_BASIC_INFO)):
            return None
        if not (_data1 := await self._get_basic_info(DEVICE_GET_BASIC_SETTINGS_KEY)):
            return None

        _LOGGER.debug("data: %s", _data)
        return self._parse_basic_info(_data, _data1)

    def _parse_basic_info(self, _data: bytes, _data1: bytes) -> dict[str, Any]:
        """Decode the basic-info connection response into a state dict."""
        battery = _data[2] & 0b01111111
        isOn = bool(_data[3] & 0b10000000)
        oscillating_horizontal = bool(_data[3] & 0b01000000)
        oscillating_vertical = bool(_data[3] & 0b00100000)
        oscillating = oscillating_horizontal or oscillating_vertical
        _mode = _data[8] & 0b00000111
        mode_enum = self._mode_enum
        max_mode = max(m.value for m in mode_enum)
        mode = mode_enum(_mode).name.lower() if 1 <= _mode <= max_mode else None
        speed = _data[9]
        firmware = _data1[2] / 10.0

        info: dict[str, Any] = {
            "battery": battery,
            "isOn": isOn,
            "oscillating": oscillating,
            "oscillating_horizontal": oscillating_horizontal,
            "oscillating_vertical": oscillating_vertical,
            "mode": mode,
            "speed": speed,
            "firmware": firmware,
        }
        # Night light is only meaningful for models that expose it. Copy from
        # the latest advertisement parse if the parser put it there.
        night_light = self._get_adv_value("nightLight")
        if night_light is not None:
            info["nightLight"] = night_light
        return info

    async def _get_basic_info(self, cmd: str) -> bytes | None:
        """Return basic info of device."""
        _data = await self._send_command(key=cmd, retry=self._retry_count)

        if _data in (b"\x07", b"\x00"):
            _LOGGER.error("Unsuccessful, please try again")
            return None

        return _data

    @update_after_operation
    async def set_preset_mode(self, preset_mode: str) -> bool:
        """Send command to set fan preset_mode."""
        result = await self._send_command(self._command_set_mode[preset_mode])
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_percentage(self, percentage: int) -> bool:
        """Send command to set fan percentage."""
        result = await self._send_command(f"{COMMAND_SET_PERCENTAGE}{percentage:02X}")
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_oscillation(self, oscillating: bool) -> bool:
        """Send command to set fan oscillation"""
        cmd = (
            self._command_start_oscillation
            if oscillating
            else self._command_stop_oscillation
        )
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_horizontal_oscillation(self, oscillating: bool) -> bool:
        """Send command to set fan horizontal (left-right) oscillation only."""
        cmd = (
            COMMAND_START_HORIZONTAL_OSCILLATION
            if oscillating
            else COMMAND_STOP_HORIZONTAL_OSCILLATION
        )
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_vertical_oscillation(self, oscillating: bool) -> bool:
        """Send command to set fan vertical (up-down) oscillation only."""
        cmd = (
            COMMAND_START_VERTICAL_OSCILLATION
            if oscillating
            else COMMAND_STOP_VERTICAL_OSCILLATION
        )
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    def get_current_percentage(self) -> Any:
        """Return cached percentage."""
        return self._get_adv_value("speed")

    def is_on(self) -> bool | None:
        """Return fan state from cache."""
        return self._get_adv_value("isOn")

    def get_oscillating_state(self) -> Any:
        """Return cached oscillating."""
        return self._get_adv_value("oscillating")

    def get_horizontal_oscillating_state(self) -> Any:
        """Return cached horizontal (left-right) oscillating state."""
        return self._get_adv_value("oscillating_horizontal")

    def get_vertical_oscillating_state(self) -> Any:
        """Return cached vertical (up-down) oscillating state."""
        return self._get_adv_value("oscillating_vertical")

    def get_current_mode(self) -> Any:
        """Return cached mode."""
        return self._get_adv_value("mode")


class SwitchbotStandingFan(SwitchbotFan):
    """Representation of a Switchbot Standing Fan (FAN2)."""

    _mode_enum: ClassVar[type[Enum]] = StandingFanMode
    _command_set_mode: ClassVar[dict[str, str]] = COMMAND_SET_STANDING_FAN_MODE
    _command_start_oscillation: ClassVar[str] = COMMAND_START_OSCILLATION_ALL_AXES
    _command_stop_oscillation: ClassVar[str] = COMMAND_STOP_OSCILLATION_ALL_AXES

    def _parse_basic_info(self, _data: bytes, _data1: bytes) -> dict[str, Any]:
        """Add the Standing-Fan-only fields to the basic-info response."""
        info = super()._parse_basic_info(_data, _data1)
        # Sweep angle as the raw device byte: horizontal is the angle in degrees
        # (30/60/90); vertical encodes 90 as 95 (see VerticalOscillationAngle).
        info["oscillating_horizontal_angle"] = _data[4]
        info["oscillating_vertical_angle"] = _data[6]
        info["charging"] = bool(_data[2] & 0b10000000)
        info["child_lock"] = bool(_data[3] & 0b00000001)
        info["display"] = bool(_data[3] & 0b00000010)
        # bit 4 = horizontal axis, bit 3 = vertical; the app toggles both at once.
        info["auto_recenter"] = bool(_data[3] & 0b00011000)
        if len(_data) > 10:
            info["sound"] = bool(_data[10] & 0b01111111)
        return info

    @update_after_operation
    async def set_horizontal_oscillation_angle(
        self, angle: HorizontalOscillationAngle | int
    ) -> bool:
        """Set horizontal oscillation angle (30 / 60 / 90 degrees)."""
        value = HorizontalOscillationAngle(angle).value
        cmd = f"{COMMAND_SET_OSCILLATION_PARAMS}{value:02X}FFFFFF"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_vertical_oscillation_angle(
        self, angle: VerticalOscillationAngle | int
    ) -> bool:
        """
        Set vertical oscillation angle (30 / 60 / 90 degrees).

        The device uses a different byte encoding on the vertical axis than
        on the horizontal one: 90° maps to byte 0x5F (95), not 0x5A (90),
        which the firmware interprets as an axis halt. Use
        `VerticalOscillationAngle` (or the raw byte values 30 / 60 / 95).
        """
        value = VerticalOscillationAngle(angle).value
        cmd = f"{COMMAND_SET_OSCILLATION_PARAMS}FFFF{value:02X}FF"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_night_light(self, state: NightLightState | int) -> bool:
        """Set night-light state (LEVEL_1, LEVEL_2, OFF)."""
        value = NightLightState(state).value
        cmd = f"{COMMAND_SET_NIGHT_LIGHT}{value:02X}FFFF"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_child_lock(self, enabled: bool) -> bool:
        """Enable or disable the child lock."""
        cmd = f"{COMMAND_SET_CHILD_LOCK}{'01' if enabled else '02'}"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_display(self, enabled: bool) -> bool:
        """Turn the front display (LED) on or off."""
        cmd = f"{COMMAND_SET_DISPLAY_LIGHT}{'01' if enabled else '02'}FFFF"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_sound(self, enabled: bool) -> bool:
        """Turn the key tone (buzzer) on or off."""
        cmd = f"{COMMAND_SET_SOUND}{'64' if enabled else '00'}"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    @update_after_operation
    async def set_auto_recenter(self, enabled: bool) -> bool:
        """Enable or disable auto return-to-center on both axes."""
        cmd = f"{COMMAND_SET_AUTO_RECENTER}{'0101' if enabled else '0202'}"
        result = await self._send_command(cmd)
        return self._check_command_result(result, 0, {1})

    def get_horizontal_oscillation_angle(self) -> int | None:
        """Return cached horizontal oscillation angle (raw device byte)."""
        return self._get_adv_value("oscillating_horizontal_angle")

    def get_vertical_oscillation_angle(self) -> int | None:
        """Return cached vertical oscillation angle (raw device byte; 90° = 95)."""
        return self._get_adv_value("oscillating_vertical_angle")

    def get_night_light_state(self) -> int | None:
        """Return cached night light state."""
        return self._get_adv_value("nightLight")

    def is_charging(self) -> bool | None:
        """Return cached charging state."""
        return self._get_adv_value("charging")

    def get_child_lock(self) -> bool | None:
        """Return cached child-lock state."""
        return self._get_adv_value("child_lock")

    def get_display(self) -> bool | None:
        """Return cached front-display (LED) state."""
        return self._get_adv_value("display")

    def get_sound(self) -> bool | None:
        """Return cached key-tone (buzzer) state."""
        return self._get_adv_value("sound")

    def get_auto_recenter(self) -> bool | None:
        """Return cached auto-recenter (return-to-center) state."""
        return self._get_adv_value("auto_recenter")



from switchbot.adv_parsers.hub2 import calculate_light_intensity


def test_calculate_light_intensity():
    """Test calculating light intensity from Hub 2 light level."""
    # Test valid inputs
    assert calculate_light_intensity(1) == 0
    assert calculate_light_intensity(2) == 10
    assert calculate_light_intensity(10) == 90
    assert calculate_light_intensity(15) == 510
    assert calculate_light_intensity(21) == 1091

    # Test invalid inputs
    assert calculate_light_intensity(0) == 0
    assert calculate_light_intensity(22) == 0
    assert calculate_light_intensity(-1) == 0
    assert calculate_light_intensity(3.5) == 0
    assert calculate_light_intensity(None) == 0



from unittest.mock import AsyncMock, MagicMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement, SwitchbotModel
from switchbot.const.light import ColorMode
from switchbot.devices import light_strip
from switchbot.devices.base_light import SwitchbotBaseLight
from switchbot.devices.device import SwitchbotOperationError

from . import (
    CANDLE_WARMER_LAMP_INFO,
    FLOOR_LAMP_INFO,
    PERMANENT_OUTDOOR_LIGHT_INFO,
    RGBIC_NEON_LIGHT_INFO,
    RGBICWW_CEILING_LIGHT_INFO,
    RGBICWW_FLOOR_LAMP_INFO,
    RGBICWW_STRIP_LIGHT_INFO,
    STRIP_LIGHT_3_INFO,
)
from .test_adv_parser import AdvTestCase, generate_ble_device

ALL_LIGHT_CASES = [
    (STRIP_LIGHT_3_INFO, light_strip.SwitchbotStripLight3),
    (FLOOR_LAMP_INFO, light_strip.SwitchbotStripLight3),
    (CANDLE_WARMER_LAMP_INFO, light_strip.SwitchbotCandleWarmerLamp),
    (RGBICWW_STRIP_LIGHT_INFO, light_strip.SwitchbotRgbicLight),
    (RGBICWW_FLOOR_LAMP_INFO, light_strip.SwitchbotRgbicLight),
    (PERMANENT_OUTDOOR_LIGHT_INFO, light_strip.SwitchbotPermanentOutdoorLight),
    (RGBIC_NEON_LIGHT_INFO, light_strip.SwitchbotRgbicNeonLight),
    (RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight),
]

# RGB + color-temp devices. Excludes brightness-only lights (CWL) and
# RGB-only lights (RGBIC Neon) whose color_modes differ from the rest.
RGB_LIGHT_CASES = [
    case
    for case in ALL_LIGHT_CASES
    if case[1]
    not in (light_strip.SwitchbotCandleWarmerLamp, light_strip.SwitchbotRgbicNeonLight)
]


@pytest.fixture(params=RGB_LIGHT_CASES)
def device_case(request):
    return request.param


@pytest.fixture
def expected_effects(device_case):
    adv_info, _dev_cls = device_case
    EXPECTED = {
        SwitchbotModel.STRIP_LIGHT_3: ("christmas", "halloween", "sunset"),
        SwitchbotModel.FLOOR_LAMP: ("christmas", "halloween", "sunset"),
        SwitchbotModel.RGBICWW_STRIP_LIGHT: ("romance", "energy", "heartbeat"),
        SwitchbotModel.RGBICWW_FLOOR_LAMP: ("romance", "energy", "heartbeat"),
        SwitchbotModel.RGBICWW_CEILING_LIGHT: ("romance", "energy", "heartbeat"),
        SwitchbotModel.PERMANENT_OUTDOOR_LIGHT: ("romance", "energy", "heartbeat"),
    }
    return EXPECTED[adv_info.modelName]


def create_device_for_command_testing(
    adv_info: AdvTestCase,
    dev_cls: type[SwitchbotBaseLight],
    init_data: dict | None = None,
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = dev_cls(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff", model=adv_info.modelName
    )
    device.update_from_advertisement(
        make_advertisement_data(ble_device, adv_info, init_data)
    )
    device._send_command = AsyncMock()
    device._check_command_result = MagicMock()
    device.update = AsyncMock()
    return device


def make_advertisement_data(
    ble_device: BLEDevice, adv_info: AdvTestCase, init_data: dict | None = None
):
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": adv_info.service_data,
            "data": adv_info.data | init_data,
            "isEncrypted": False,
            "model": adv_info.model,
            "modelFriendlyName": adv_info.modelFriendlyName,
            "modelName": adv_info.modelName,
        },
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_default_info(device_case, expected_effects):
    """Test default initialization of the strip light."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    assert device.rgb is None

    device._state = {"r": 30, "g": 0, "b": 0, "cw": 3200}

    assert device.is_on() is True
    assert device.on is True
    assert device.color_mode == ColorMode.RGB
    assert device.color_modes == {ColorMode.RGB, ColorMode.COLOR_TEMP}
    assert device.rgb == (30, 0, 0)
    assert device.color_temp == 3200
    assert device.brightness == adv_info.data["brightness"]
    assert device.min_temp == 2700
    assert device.max_temp == 6500
    # Check that effect list contains expected lowercase effect names
    effect_list = device.get_effect_list
    assert effect_list is not None
    assert all(effect.islower() for effect in effect_list)
    # Verify some known effects are present
    for effect in expected_effects:
        assert effect in effect_list


@pytest.mark.asyncio
async def test_rgbic_neon_light_info() -> None:
    """Test color_mode / color_modes on SwitchbotRgbicNeonLight (RGB only)."""
    adv_info, dev_cls = RGBIC_NEON_LIGHT_INFO, light_strip.SwitchbotRgbicNeonLight
    device = create_device_for_command_testing(adv_info, dev_cls)
    assert device.color_mode == ColorMode.RGB
    assert device.color_modes == {ColorMode.RGB}


@pytest.mark.asyncio
async def test_candle_warmer_lamp_info() -> None:
    """Test default initialization of the candle warmer lamp."""
    adv_info, dev_cls = CANDLE_WARMER_LAMP_INFO, light_strip.SwitchbotCandleWarmerLamp
    device = create_device_for_command_testing(adv_info, dev_cls)
    assert device.rgb is None
    assert device.is_on() is True
    assert device.on is True
    assert device.color_mode == ColorMode.BRIGHTNESS
    assert device.color_modes == {
        ColorMode.BRIGHTNESS,
    }
    assert device.brightness == adv_info.data["brightness"]
    assert device.get_effect_list is None


@pytest.mark.asyncio
async def test_candle_warmer_lamp_unsupported_operations() -> None:
    """Test that RGB/color-temp/effect operations are not supported on CWL."""
    device = create_device_for_command_testing(
        CANDLE_WARMER_LAMP_INFO, light_strip.SwitchbotCandleWarmerLamp
    )
    with pytest.raises(
        SwitchbotOperationError,
        match="does not support this functionality",
    ):
        await device.set_rgb(100, 255, 128, 64)
    with pytest.raises(
        SwitchbotOperationError,
        match="does not support this functionality",
    ):
        await device.set_color_temp(100, 4000)
    with pytest.raises(SwitchbotOperationError, match="not supported"):
        await device.set_effect("sunset")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("basic_info", "version_info"), [(True, False), (False, True), (False, False)]
)
@pytest.mark.parametrize(("adv_info", "dev_cls"), ALL_LIGHT_CASES)
async def test_get_basic_info_returns_none(
    basic_info, version_info, adv_info, dev_cls
) -> None:
    """Test that get_basic_info returns None if no data is available."""
    device = create_device_for_command_testing(adv_info, dev_cls)

    device._send_command = AsyncMock(side_effect=[version_info, basic_info])
    device._check_command_result = MagicMock(
        side_effect=[bool(version_info), bool(basic_info)]
    )

    assert await device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "device_case",
    [
        case
        for case in RGB_LIGHT_CASES
        if case[1] is not light_strip.SwitchbotRgbicwwCeilingLight
    ],
)
@pytest.mark.parametrize(
    ("info_data", "result"),
    [
        (
            {
                "basic_info": b"\x01\x00<\xff\x00\xd8\x00\x19d\x00\x03",
                "version_info": b"\x01\x01\n",
            },
            [False, 60, 255, 0, 216, 6500, 3, 1.0],
        ),
        (
            {
                "basic_info": b"\x01\x80NK\xff:\x00\x19d\xff\x02",
                "version_info": b"\x01\x01\n",
            },
            [True, 78, 75, 255, 58, 6500, 2, 1.0],
        ),
        (
            {
                "basic_info": b"\x01\x80$K\xff:\x00\x13\xf9\xff\x06",
                "version_info": b"\x01\x01\n",
            },
            [True, 36, 75, 255, 58, 5113, 6, 1.0],
        ),
    ],
)
async def test_strip_light_get_basic_info(info_data, result, device_case):
    """Test getting basic info from the strip light."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    device._send_command = AsyncMock(
        side_effect=[info_data["version_info"], info_data["basic_info"]]
    )
    device._check_command_result = MagicMock(side_effect=[True, True])
    info = await device.get_basic_info()

    assert info["isOn"] is result[0]
    assert info["brightness"] == result[1]
    assert info["r"] == result[2]
    assert info["g"] == result[3]
    assert info["b"] == result[4]
    assert info["cw"] == result[5]
    assert info["color_mode"] == result[6]
    assert info["firmware"] == result[7]


@pytest.mark.asyncio
async def test_set_color_temp(device_case):
    """Test setting color temperature."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    await device.set_color_temp(50, 3000)

    device._send_command.assert_called_with(
        device._set_color_temp_command.format("320BB8")
    )


@pytest.mark.asyncio
async def test_turn_on(device_case):
    """Test turning on the strip light."""
    init_data = {"isOn": True}
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls, init_data)

    await device.turn_on()

    device._send_command.assert_called_with(device._turn_on_command)

    assert device.is_on() is True


@pytest.mark.asyncio
async def test_turn_off(device_case):
    """Test turning off the strip light."""
    init_data = {"isOn": False}
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls, init_data)

    await device.turn_off()

    device._send_command.assert_called_with(device._turn_off_command)

    assert device.is_on() is False


@pytest.mark.asyncio
async def test_set_brightness(device_case):
    """Test setting brightness."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    await device.set_brightness(75)

    device._send_command.assert_called_with(device._set_brightness_command.format("4B"))


@pytest.mark.asyncio
async def test_set_rgb(device_case):
    """Test setting RGB values."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    await device.set_rgb(100, 255, 128, 64)

    device._send_command.assert_called_with(device._set_rgb_command.format("64FF8040"))


@pytest.mark.asyncio
async def test_set_effect_with_invalid_effect(device_case):
    """Test setting an invalid effect."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    with pytest.raises(
        SwitchbotOperationError, match="Effect invalid_effect not supported"
    ):
        await device.set_effect("invalid_effect")


@pytest.mark.asyncio
async def test_set_effect_with_valid_effect(device_case):
    """Test setting a valid effect."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)
    device._send_multiple_commands = AsyncMock()

    await device.set_effect("christmas")

    device._send_multiple_commands.assert_called_with(device._effect_dict["christmas"])

    assert device.get_effect() == "christmas"


@pytest.mark.asyncio
async def test_effect_list_contains_lowercase_names(device_case, expected_effects):
    """Test that all effect names in get_effect_list are lowercase."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    effect_list = device.get_effect_list

    assert effect_list is not None, "Effect list should not be None"
    # All effect names should be lowercase
    for effect_name in effect_list:
        assert effect_name.islower(), f"Effect name '{effect_name}' is not lowercase"
    # Verify some known effects are present
    for expected_effect in expected_effects:
        assert expected_effect in effect_list, (
            f"Expected effect '{expected_effect}' not found"
        )


@pytest.mark.asyncio
async def test_set_effect_normalizes_case(device_case):
    """Test that set_effect normalizes effect names to lowercase."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)
    device._send_multiple_commands = AsyncMock()

    # Test various case combinations
    test_cases = ["CHRISTMAS", "Christmas", "ChRiStMaS", "christmas"]

    for test_effect in test_cases:
        await device.set_effect(test_effect)
        # Should always work regardless of case
        device._send_multiple_commands.assert_called()
        assert device.get_effect() == test_effect  # Stored as provided


@pytest.mark.parametrize(
    ("dev_cls", "expected_model"),
    [
        (light_strip.SwitchbotStripLight3, SwitchbotModel.STRIP_LIGHT_3),
        (light_strip.SwitchbotRgbicLight, SwitchbotModel.RGBICWW_STRIP_LIGHT),
        (
            light_strip.SwitchbotPermanentOutdoorLight,
            SwitchbotModel.PERMANENT_OUTDOOR_LIGHT,
        ),
        (light_strip.SwitchbotRgbicNeonLight, SwitchbotModel.RGBIC_NEON_ROPE_LIGHT),
        (light_strip.SwitchbotCandleWarmerLamp, SwitchbotModel.CANDLE_WARMER_LAMP),
        (
            light_strip.SwitchbotRgbicwwCeilingLight,
            SwitchbotModel.RGBICWW_CEILING_LIGHT,
        ),
    ],
)
def test_default_model_classvar(dev_cls, expected_model):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = dev_cls(ble_device, "ff", "ffffffffffffffffffffffffffffffff")
    assert device._model == expected_model


def create_strip_light_device(init_data: dict | None = None):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    return light_strip.SwitchbotLightStrip(ble_device)


@pytest.mark.asyncio
async def test_strip_light_supported_color_modes():
    """Test that the strip light supports the expected color modes."""
    device = create_strip_light_device()

    assert device.color_modes == {
        ColorMode.RGB,
    }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("commands", "results", "final_result"),
    [
        (("command1", "command2"), [(b"\x01", False), (None, False)], False),
        (("command1", "command2"), [(None, False), (b"\x01", True)], True),
        (("command1", "command2"), [(b"\x01", True), (b"\x01", False)], True),
    ],
)
async def test_send_multiple_commands(commands, results, final_result, device_case):
    """Test sending multiple commands."""
    adv_info, dev_cls = device_case
    device = create_device_for_command_testing(adv_info, dev_cls)

    device._send_command = AsyncMock(side_effect=[r[0] for r in results])

    device._check_command_result = MagicMock(side_effect=[r[1] for r in results])

    result = await device._send_multiple_commands(list(commands))

    assert result is final_result


@pytest.mark.asyncio
async def test_unimplemented_color_mode():
    class TestDevice(SwitchbotBaseLight):
        pass

    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = TestDevice(ble_device)

    with pytest.raises(NotImplementedError):
        _ = device.color_mode


@pytest.mark.asyncio
async def test_exception_with_wrong_model():
    class TestDevice(SwitchbotBaseLight):
        def __init__(self, device: BLEDevice, model: str = "unknown") -> None:
            super().__init__(device, model=model)

    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = TestDevice(ble_device)

    with pytest.raises(
        SwitchbotOperationError,
        match="Current device aa:bb:cc:dd:ee:ff does not support this functionality",
    ):
        await device.set_rgb(100, 255, 128, 64)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("color_mode_value", "expected_color_mode"),
    [
        (1, ColorMode.EFFECT),  # SEGMENTED
        (2, ColorMode.RGB),
        (3, ColorMode.EFFECT),  # SCENE
        (4, ColorMode.EFFECT),  # MUSIC
        (5, ColorMode.EFFECT),  # CONTROLLER
        (6, ColorMode.COLOR_TEMP),
        (7, ColorMode.EFFECT),  # EFFECT (RGBIC-specific)
        (10, ColorMode.OFF),  # UNKNOWN
    ],
)
async def test_permanent_outdoor_light_color_mode(
    color_mode_value, expected_color_mode
):
    """Test that POL correctly handles all RGBICStripLightColorMode values including EFFECT (7)."""
    device = create_device_for_command_testing(
        PERMANENT_OUTDOOR_LIGHT_INFO,
        light_strip.SwitchbotPermanentOutdoorLight,
        init_data={"color_mode": color_mode_value},
    )
    assert device.color_mode == expected_color_mode


@pytest.mark.asyncio
async def test_rgbicww_ceiling_light_main_sub_light_state() -> None:
    """Test the main (warm-white) sub-light state surface."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    assert device.is_main_on is True
    assert device.main_brightness == 32


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("method", "expected_command_attr"),
    [
        ("turn_on_main", "_turn_on_main_command"),
        ("turn_off_main", "_turn_off_main_command"),
        ("turn_on_color", "_turn_on_color_command"),
        ("turn_off_color", "_turn_off_color_command"),
    ],
)
async def test_rgbicww_ceiling_light_sub_light_power(
    method: str, expected_command_attr: str
) -> None:
    """Test the per-sub-light power commands."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    await getattr(device, method)()
    device._send_command.assert_called_with(getattr(device, expected_command_attr))


@pytest.mark.asyncio
async def test_rgbicww_ceiling_light_set_main_brightness() -> None:
    """Test setting main (warm-white) sub-light brightness."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    await device.set_main_brightness(75)
    device._send_command.assert_called_with(
        device._set_main_brightness_command.format("4B")
    )


@pytest.mark.asyncio
async def test_rgbicww_ceiling_light_set_main_color_temp() -> None:
    """Test setting main (warm-white) sub-light color temperature."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    await device.set_main_color_temp(3000)
    device._send_command.assert_called_with(
        device._set_main_color_temp_command.format("0BB8")
    )


@pytest.mark.asyncio
async def test_rgbicww_ceiling_light_color_modes() -> None:
    """Test the supported color modes set."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    assert device.color_modes == {ColorMode.RGB, ColorMode.COLOR_TEMP}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("color_mode_value", "expected_color_mode"),
    [
        (1, ColorMode.EFFECT),  # SEGMENTED
        (2, ColorMode.RGB),  # COLOR
        (3, ColorMode.EFFECT),  # SCENE
        (4, ColorMode.EFFECT),  # MUSIC
        (5, ColorMode.EFFECT),  # CONTROLLER
        (6, ColorMode.COLOR_TEMP),  # WARMWHITE
        (7, ColorMode.EFFECT),  # EFFECT
        (10, ColorMode.OFF),  # UNKNOWN
    ],
)
async def test_rgbicww_ceiling_light_color_mode(
    color_mode_value, expected_color_mode
) -> None:
    """Test color_mode mapping for the ceiling light."""
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO,
        light_strip.SwitchbotRgbicwwCeilingLight,
        init_data={"color_mode": color_mode_value},
    )
    assert device.color_mode == expected_color_mode


@pytest.mark.asyncio
async def test_rgbicww_ceiling_light_get_basic_info_with_main_state() -> None:
    """
    Test that the ceiling light's get_basic_info reads RGB over GATT.

    Power/brightness/mode/color-temp come from the advertisement, not GATT,
    so they are not part of the get_basic_info result.
    """
    device = create_device_for_command_testing(
        RGBICWW_CEILING_LIGHT_INFO, light_strip.SwitchbotRgbicwwCeilingLight
    )
    device._send_command = AsyncMock(
        side_effect=[b"\x01\x01\n", b"\x01\x80NK\xff:\xa0\x19d\xff\x02"]
    )
    device._check_command_result = MagicMock(side_effect=[True, True])

    info = await device.get_basic_info()

    assert info["r"] == 75
    assert info["g"] == 255
    assert info["b"] == 58
    assert info["firmware"] == 1.0
    # Power state must not be sourced from GATT (advertisement is authoritative).
    assert "isOn" not in info
    assert "main_isOn" not in info



from unittest.mock import AsyncMock, MagicMock

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement
from switchbot.const.climate import ClimateAction, ClimateMode
from switchbot.const.climate import SmartThermostatRadiatorMode as STRMode
from switchbot.devices.device import SwitchbotOperationError
from switchbot.devices.smart_thermostat_radiator import (
    COMMAND_SET_MODE,
    COMMAND_SET_TEMP,
    SwitchbotSmartThermostatRadiator,
)

from . import SMART_THERMOSTAT_RADIATOR_INFO
from .test_adv_parser import AdvTestCase, generate_ble_device


def create_device_for_command_testing(
    adv_info: AdvTestCase,
    init_data: dict | None = None,
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = SwitchbotSmartThermostatRadiator(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff", model=adv_info.modelName
    )
    device.update_from_advertisement(
        make_advertisement_data(ble_device, adv_info, init_data)
    )
    device._send_command = AsyncMock()
    device._check_command_result = MagicMock()
    device.update = AsyncMock()
    return device


def make_advertisement_data(
    ble_device: BLEDevice, adv_info: AdvTestCase, init_data: dict | None = None
):
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": adv_info.service_data,
            "data": adv_info.data | init_data,
            "isEncrypted": False,
            "model": adv_info.model,
            "modelFriendlyName": adv_info.modelFriendlyName,
            "modelName": adv_info.modelName,
        }
        | init_data,
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_default_info() -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)

    assert device.min_temperature == 5.0
    assert device.max_temperature == 35.0
    assert device.preset_mode == STRMode.MANUAL.lname
    assert device.preset_modes == STRMode.get_modes()
    assert device.hvac_mode == ClimateMode.HEAT
    assert device.hvac_modes == {ClimateMode.OFF, ClimateMode.HEAT}
    assert device.hvac_action == ClimateAction.HEATING
    assert device.target_temperature == 35.0
    assert device.current_temperature == 28.0
    assert device.door_open() is False


@pytest.mark.asyncio
async def test_default_info_with_off_mode() -> None:
    device = create_device_for_command_testing(
        SMART_THERMOSTAT_RADIATOR_INFO, {"mode": STRMode.OFF.lname, "isOn": False}
    )
    assert device.hvac_action == ClimateAction.OFF


@pytest.mark.parametrize(
    ("mode", "expected_command"),
    [
        (ClimateMode.OFF, "570100"),
        (ClimateMode.HEAT, COMMAND_SET_MODE[STRMode.COMFORT.lname]),
    ],
)
@pytest.mark.asyncio
async def test_set_hvac_mode_commands(mode, expected_command) -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)

    await device.set_hvac_mode(mode)
    device._send_command.assert_awaited_with(expected_command)


@pytest.mark.parametrize(
    ("preset_mode", "expected_command"),
    [
        (STRMode.SCHEDULE.lname, COMMAND_SET_MODE[STRMode.SCHEDULE.lname]),
        (STRMode.MANUAL.lname, COMMAND_SET_MODE[STRMode.MANUAL.lname]),
        (STRMode.OFF.lname, COMMAND_SET_MODE[STRMode.OFF.lname]),
        (STRMode.ECO.lname, COMMAND_SET_MODE[STRMode.ECO.lname]),
        (STRMode.COMFORT.lname, COMMAND_SET_MODE[STRMode.COMFORT.lname]),
        (STRMode.BOOST.lname, COMMAND_SET_MODE[STRMode.BOOST.lname]),
    ],
)
@pytest.mark.asyncio
async def test_set_preset_mode_commands(preset_mode, expected_command) -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)

    await device.set_preset_mode(preset_mode)
    device._send_command.assert_awaited_with(expected_command)


@pytest.mark.asyncio
async def test_set_target_temperature_command() -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)

    await device.set_target_temperature(22.5)
    device._send_command.assert_awaited_with(
        COMMAND_SET_TEMP[STRMode.MANUAL.lname].format(temp=225)
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("mode", "match"),
    [
        (STRMode.OFF.lname, "Cannot set temperature when mode is OFF."),
        (STRMode.BOOST.lname, "Boost mode defaults to max temperature."),
    ],
)
async def test_set_target_temperature_with_invalid_mode(mode, match) -> None:
    device = create_device_for_command_testing(
        SMART_THERMOSTAT_RADIATOR_INFO, {"mode": mode}
    )

    with pytest.raises(SwitchbotOperationError, match=match):
        await device.set_target_temperature(22.5)


@pytest.mark.asyncio
async def test_get_basic_info_none() -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)
    device._get_basic_info = AsyncMock(return_value=None)

    assert await device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("basic_info", "result"),
    [
        (
            b"\x01d\x08>\x14\x80\xe6\x00(\x82\xbe\x00T\x00\x82\x00\x00",
            [
                100,
                0.8,
                62,
                "off",
                "comfort",
                23.0,
                4.0,
                13.0,
                19.0,
                0,
                False,
                13.0,
                False,
            ],
        ),
        (
            b"\x01d\x08>#\x80\xf0\x00(\x82\xbe\x00T\x00\x82\x00\x00",
            [
                100,
                0.8,
                62,
                "comfort",
                "eco",
                24.0,
                4.0,
                13.0,
                19.0,
                0,
                False,
                13.0,
                False,
            ],
        ),
    ],
)
async def test_get_basic_info_parsing(basic_info, result) -> None:
    device = create_device_for_command_testing(SMART_THERMOSTAT_RADIATOR_INFO)
    device._get_basic_info = AsyncMock(return_value=basic_info)

    info = await device.get_basic_info()
    assert info["battery"] == result[0]
    assert info["firmware"] == result[1]
    assert info["hardware"] == result[2]
    assert info["last_mode"] == result[3]
    assert info["mode"] == result[4]
    assert info["temperature"] == result[5]
    assert info["manual_target_temp"] == result[6]
    assert info["comfort_target_temp"] == result[7]
    assert info["economic_target_temp"] == result[8]
    assert info["fast_heat_time"] == result[9]
    assert info["child_lock"] == result[10]
    assert info["target_temp"] == result[11]
    assert info["door_open"] == result[12]


def test_default_model_classvar():
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = SwitchbotSmartThermostatRadiator(
        ble_device, "ff", "ffffffffffffffffffffffffffffffff"
    )
    assert device._model == SMART_THERMOSTAT_RADIATOR_INFO.modelName


@pytest.mark.asyncio
async def test_thermostat_idle_action() -> None:
    """Test TRV action transitions exactly at the 0.5°C hysteresis boundary."""
    # Case 1: Room temperature is exactly at target + 0.5 -> Should be IDLE
    device_at_boundary = create_device_for_command_testing(
        SMART_THERMOSTAT_RADIATOR_INFO,
        {"target_temperature": 20.0, "temperature": 20.5},
    )
    assert device_at_boundary.hvac_action == ClimateAction.IDLE

    # Case 2: Room is warm but within the 0.5 chattering band -> Should still be HEATING
    device_in_band = create_device_for_command_testing(
        SMART_THERMOSTAT_RADIATOR_INFO,
        {"target_temperature": 20.0, "temperature": 20.4},
    )
    assert device_in_band.hvac_action == ClimateAction.HEATING



"""Hubmini matter parser."""

from __future__ import annotations

from typing import Any

from ..helpers import celsius_to_fahrenheit


def process_hubmini_matter(
    data: bytes | None, mfr_data: bytes | None
) -> dict[str, Any]:
    """Process Hubmini matter sensor manufacturer data."""
    temp_data = None

    if mfr_data and len(mfr_data) >= 16:
        temp_data = mfr_data[13:16]

    if not temp_data:
        return {}

    _temp_sign = 1 if temp_data[1] & 0b10000000 else -1
    _temp_c = _temp_sign * (
        (temp_data[1] & 0b01111111) + ((temp_data[0] & 0b00001111) / 10)
    )
    _temp_f = celsius_to_fahrenheit(_temp_c)
    _temp_f = (_temp_f * 10) / 10
    humidity = temp_data[2] & 0b01111111

    if _temp_c == 0 and humidity == 0:
        return {}

    return {
        "temp": {"c": _temp_c, "f": _temp_f},
        "temperature": _temp_c,
        "fahrenheit": bool(temp_data[2] & 0b10000000),
        "humidity": humidity,
    }



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from bleak.backends.device import BLEDevice

from switchbot import SwitchBotAdvertisement, SwitchbotModel
from switchbot.const.light import ColorMode
from switchbot.devices import ceiling_light

from .test_adv_parser import generate_ble_device


def create_device_for_command_testing(
    init_data: dict | None = None, model: SwitchbotModel = SwitchbotModel.CEILING_LIGHT
):
    ble_device = generate_ble_device("aa:bb:cc:dd:ee:ff", "any")
    device = ceiling_light.SwitchbotCeilingLight(ble_device, model=model)
    device.update_from_advertisement(make_advertisement_data(ble_device, init_data))
    device._send_command = AsyncMock()
    device._check_command_result = MagicMock()
    device.update = AsyncMock()
    return device


def make_advertisement_data(ble_device: BLEDevice, init_data: dict | None = None):
    """Set advertisement data with defaults."""
    if init_data is None:
        init_data = {}

    return SwitchBotAdvertisement(
        address="aa:bb:cc:dd:ee:ff",
        data={
            "rawAdvData": b"q\x00",
            "data": {
                "brightness": 1,
                "color_mode": 1,
                "cw": 6387,
                "isOn": False,
                "sequence_number": 10,
            }
            | init_data,
            "isEncrypted": False,
            "model": b"q\x00",
            "modelFriendlyName": "Ceiling Light",
            "modelName": SwitchbotModel.CEILING_LIGHT,
        },
        device=ble_device,
        rssi=-80,
        active=True,
    )


@pytest.mark.asyncio
async def test_default_info():
    """Test default initialization of the ceiling light."""
    device = create_device_for_command_testing()

    assert device.rgb is None

    device._state = {"cw": 3200}

    assert device.is_on() is False
    assert device.on is False
    assert device.color_mode == ColorMode.COLOR_TEMP
    assert device.color_modes == {ColorMode.COLOR_TEMP}
    assert device.color_temp == 3200
    assert device.brightness == 1
    assert device.min_temp == 2700
    assert device.max_temp == 6500
    assert device.get_effect_list is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("basic_info", "version_info"), [(True, False), (False, True), (False, False)]
)
async def test_get_basic_info_returns_none(basic_info, version_info):
    device = create_device_for_command_testing()
    device._send_command = AsyncMock(side_effect=[version_info, basic_info])
    device._check_command_result = MagicMock(
        side_effect=[bool(version_info), bool(basic_info)]
    )

    assert await device.get_basic_info() is None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("info_data", "result"),
    [
        (
            {
                "basic_info": b"\x01\x80=\x0f\xa1\x00\x01",
                "version_info": b"\x01d\x15\x0f\x00\x00\x00\x00\x00\x00\x00\n\x00",
            },
            [True, 61, 4001, 0, 2.1],
        ),
        (
            {
                "basic_info": b"\x01\x80\x0e\x12B\x00\x01",
                "version_info": b"\x01d\x15\x0f\x00\x00\x00\x00\x00\x00\x00\n\x00",
            },
            [True, 14, 4674, 0, 2.1],
        ),
        (
            {
                "basic_info": b"\x01\x00\x0e\x10\x96\x00\x01",
                "version_info": b"\x01d\x15\x0f\x00\x00\x00\x00\x00\x00\x00\n\x00",
            },
            [False, 14, 4246, 0, 2.1],
        ),
    ],
)
async def test_get_basic_info(info_data, result):
    """Test getting basic info from the ceiling light."""
    device = create_device_for_command_testing()
    device._send_command = AsyncMock(
        side_effect=[info_data["version_info"], info_data["basic_info"]]
    )
    device._check_command_result = MagicMock(side_effect=[True, True])
    info = await device.get_basic_info()

    assert info["isOn"] is result[0]
    assert info["brightness"] == result[1]
    assert info["cw"] == result[2]
    assert info["color_mode"] == result[3]
    assert info["firmware"] == result[4]


@pytest.mark.asyncio
async def test_set_color_temp():
    """Test setting color temperature."""
    device = create_device_for_command_testing()

    await device.set_color_temp(50, 3000)

    device._send_command.assert_called_with(
        device._set_color_temp_command.format("320BB8")
    )


@pytest.mark.asyncio
async def test_turn_on():
    """Test turning on the ceiling light."""
    device = create_device_for_command_testing({"isOn": True})

    await device.turn_on()

    device._send_command.assert_called_with(device._turn_on_command)

    assert device.is_on() is True


@pytest.mark.asyncio
async def test_turn_off():
    """Test turning off the ceiling light."""
    device = create_device_for_command_testing({"isOn": False})

    await device.turn_off()

    device._send_command.assert_called_with(device._turn_off_command)

    assert device.is_on() is False


@pytest.mark.asyncio
async def test_set_brightness():
    """Test setting brightness."""
    device = create_device_for_command_testing()

    await device.set_brightness(75)

    device._send_command.assert_called_with(
        device._set_brightness_command.format("4B0FA1")
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("adv_value", "expected_color_mode"),
    [
        (0, ColorMode.COLOR_TEMP),
        (1, ColorMode.COLOR_TEMP),
        (4, ColorMode.EFFECT),
        (10, ColorMode.OFF),
        (None, ColorMode.OFF),
    ],
)
async def test_get_color_mode(adv_value, expected_color_mode):
    """Test getting color mode."""
    device = create_device_for_command_testing()

    with patch.object(device, "_get_adv_value", return_value=adv_value):
        assert device.color_mode == expected_color_mode
